#' Comparison of metabolite sets between dynamics clusters of different experimental conditions
#'
#' Uses the Jaccard Index to compare metabolite names between dynamics clusters of
#' different experimental conditions
#'
#' @param clusters a dataframe or containing a column specifying the metabolite
#' names to be compared and cluster IDs (column named "cluster") of
#' clusters of similar dynamics, as well as a column "condition" specifying
#' the experimental conditions
#' to be compared
#' @param metabolite column in "clusters" that specifies either metabolite name
#' or KEGG ID or some other identifier
#'
#' @return a list holding a 1) dataframe of Jaccard indices between clusters
#'
#' @export
#'
#' @seealso Visualization of metabolite similarity [heatmap_metabolites()]/
#' compare dynamics of clusters [compare_dynamics()]
#'
#' @examples
#' data("cluster")
#' comparison <- compare_metabolites(
#'   clusters = cluster
#' )
#' head(comparison)
compare_metabolites <- function(clusters, metabolite = "metabolite") {
  # Input checks
  if (!is.data.frame(clusters)) stop("'clusters' must be a dataframe")
  if (!is.character(metabolite)) stop("'metabolite' must be a character vector")
  if (!all(c("cluster") %in% colnames(clusters))) {
    stop("'clusters' must contain a column named 'cluster'")
  }
  if (!all(c(metabolite) %in% colnames(clusters))) {
    stop("'clusters' must contain a column containing
         metabolite names as specified with metabolite= ")
  }

  # Extract unique condition-cluster combinations
  unique_combinations <- unique(clusters[, c("condition", "cluster")])
  n_combinations <- nrow(unique_combinations)

  # Generate upper triangle of pairwise combinations (i.e., no self-comparisons)
  comparison_pairs <- combn(seq_len(n_combinations), 2)

  # Assign names to distances
  comparison_names <- apply(comparison_pairs, 2, function(idx) {
    paste(
      paste(unique_combinations$condition[idx[1]], unique_combinations$cluster[idx[1]], sep = "_"),
      "vs",
      paste(unique_combinations$condition[idx[2]], unique_combinations$cluster[idx[2]], sep = "_")
    )
  })

  # Function to calculate Jaccard index
  calculate_jaccard <- function(group_a, group_b) {
    .similarity(group_a, group_b) # Assuming .similarity calculates Jaccard Index
  }

  # Prepare metabolite lists for all combinations
  metabolite_lists <- lapply(seq_len(n_combinations), function(i) {
    subset <- clusters[clusters$condition == unique_combinations[i, "condition"] &
      clusters$cluster == unique_combinations[i, "cluster"], metabolite]
    unique(subset)
  })

  # Vectorized computation of Jaccard indices
  jaccard_indices <- apply(comparison_pairs, 2, function(idx) {
    calculate_jaccard(metabolite_lists[[idx[1]]], metabolite_lists[[idx[2]]])
  })

  # Create result data frame
  result <- data.frame(
    comparison = comparison_names,
    cluster_a = apply(comparison_pairs, 2, function(idx) {
      paste(unique_combinations[idx[1], "condition"], unique_combinations[idx[1], "cluster"], sep = "_")
    }),
    cluster_b = apply(comparison_pairs, 2, function(idx) {
      paste(unique_combinations[idx[2], "condition"], unique_combinations[idx[2], "cluster"], sep = "_")
    }),
    Jaccard = jaccard_indices
  )

  return(result)
}
